from .config import parsed_file, packages, commands, Command
import pytest


EXAMPLE_CONFIG = """\
[upload]
precondition = test -e upload.py
command = bin/python upload.py

[test]
command = bin/pytest

[run]
command = bin/run

[run-through]
ignore-exit-code = true
command = bin/run

[packages]
foo.a
bar.x"""


@pytest.fixture('function')
def config_file(tmpdir):
    """Create a config file with the given content."""
    def config(content=EXAMPLE_CONFIG):
        data = tmpdir.join('config.ini')
        data.write(content)
        return parsed_file(data.open())
    return config


def test_config__packages__2(config_file):
    """It raises a RuntimeError if the [packages] section is missing."""
    with pytest.raises(RuntimeError) as err:
        packages(config_file('[test]\ncommand=bin/test'))
    assert ('Missing the section [packages] in the config file.' ==
            str(err.value))


def test_config__commands__1(config_file):
    """It returns the selected commands from the config file."""
    assert ((Command('bin/pytest'),
             Command('bin/python upload.py',
                     precondition='test -e upload.py'),
             Command('bin/run', ignore_exit_code=True)) ==
            commands(config_file(), ['test', 'upload', 'run-through']))


def test_config__commands__3(config_file):
    """It raises a `RuntimeError` if a requested command is missing."""
    file = config_file()
    with pytest.raises(RuntimeError) as err:
        commands(file, ['run', 'stop'])
    assert ("Section [commands] in the config file does not contain the key "
            "'stop' you requested to execute." ==
            str(err.value))


def test_config__Command____repr____1():
    """It contains the name of the wrapped command."""
    cmd = Command('bin/test -v')
    assert "<Command 'bin/test -v'>" == repr(cmd)


def test_config__Command____repr____2():
    """It contains the name of the precondition."""
    cmd = Command('bin/test -v', 'test -e bin/test', ignore_exit_code=True)
    assert ("<Command 'bin/test -v' if 'test -e bin/test' ignore-exit-code>" ==
            repr(cmd))


def test_config__Command____eq____1():
    """It is equal to an identical command."""
    cmd1 = Command('bin/test')
    cmd2 = Command('bin/test')
    assert cmd1 == cmd2


def test_config__Command____ne____1():
    """It is not equal to a command with a different command."""
    cmd1 = Command('bin/test1', precondition='test -e bin/test')
    cmd2 = Command('bin/test2', precondition='test -e bin/test')
    assert cmd1 != cmd2


def test_config__Command____ne____2():
    """It is not equal to a command with a different precondition."""
    cmd1 = Command('bin/test')
    cmd2 = Command('bin/test', precondition='test -e bin/test')
    assert cmd1 != cmd2


def test_config__Command____ne____2_5():
    """It is not equal to a command with a different ignore-exit-code."""
    cmd1 = Command('bin/test')
    cmd2 = Command('bin/test', ignore_exit_code=True)
    assert cmd1 != cmd2


def test_config__Command____ne____3():
    """It is not equal to something which is not a command."""
    cmd = Command('bin/test')
    assert cmd != 'bin/test'



from .config import Command
import os.path
import pkg_resources
import re
import sys

raw_test_command = '{0.executable} setup.py -q test'.format(sys)
test_command = Command(raw_test_command)


def get_packages(*names):
    """Get relative package path.

    We use paths relative to the package name to be able to test os.chdir
    calls.
    """
    base_path = "{0}{1.sep}".format(os.getcwd(), os)
    return [
        pkg_resources.resource_filename(
            'toll', 'fixtures/{}'.format(x)).replace(base_path, '')
        for x in names]


def remove_ansi_codes(text):
    """Remove ANSI colour codes from text."""
    ansi_escape = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
    return ansi_escape.sub('', text)



try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser  # noqa

try:
    from unittest import mock
except ImportError:
    try:
        import mock  # noqa
    except ImportError:
        pass

try:
    from shutil import get_terminal_size
except ImportError:
    from backports.shutil_get_terminal_size import get_terminal_size  # noqa



from .config import Command
from .run import Runner
from .testing import get_packages, test_command, raw_test_command
from .testing import remove_ansi_codes
import re
import sys


def test_run__Runner____call____1():
    """It returns `True` on success of all packages."""
    runner = Runner([test_command], get_packages('fine', 'finetoo'))
    assert True is runner()


def test_run__Runner____call____2(capsys):
    """It prints the tests results to stdout."""
    pkgs = get_packages('fine', 'finetoo')
    runner = Runner([test_command], pkgs)
    runner()
    out, err = capsys.readouterr()
    out = remove_ansi_codes(out)
    assert 'Running {0} on {1[0]}'.format(raw_test_command, pkgs) in out
    assert 'Running {0} on {1[1]}'.format(raw_test_command, pkgs) in out
    assert out.strip().endswith('SUCCESS :-)')


def test_run__Runner____call____3(capsys):
    """It stops after the first package with a failure."""
    pkgs = get_packages('bad', 'fine')
    runner = Runner([test_command], pkgs)
    runner()
    out, err = capsys.readouterr()
    out = remove_ansi_codes(out)
    assert 'Running {0} on {1[0]}'.format(raw_test_command, pkgs) in out
    assert 'Running {0} on {1[1]}'.format(raw_test_command, pkgs) not in out
    assert out.strip().endswith('FAILURE :-(')


def test_run__Runner____call____4(capsys):
    """It runs multiple commands on each package."""
    pkgs = get_packages('fine', 'finetoo')
    runner = Runner([test_command, Command(sys.executable)], pkgs)
    runner()
    out, err = capsys.readouterr()
    out = remove_ansi_codes(out)
    assert ("""\
Running {0} on {2[0]}
Running {0} on {2[1]}
Running {1} on {2[0]}
Running {1} on {2[1]}""".format(raw_test_command, sys.executable, pkgs) ==
            "\n".join(re.findall('^Running .* on .*$', out, re.M)))
    assert out.strip().endswith('SUCCESS :-)')


def test_run__Runner____call____5(capsys):
    """It omits the command run if the precondition is not met."""
    pkgs = get_packages('bad', 'fine')
    precondition = 'test -e fine.py'
    runner = Runner([Command(raw_test_command, precondition)], pkgs)
    runner()
    out, err = capsys.readouterr()
    out = remove_ansi_codes(out)
    assert out.strip().endswith('SUCCESS :-)')
    assert 'Not running {0} on {1[0]}'.format(raw_test_command, pkgs) in out
    assert 'Precondition {0} on {1[0]} not met.'.format(
        precondition, pkgs) in out
    assert 'Running {0} on {1[1]}'.format(raw_test_command, pkgs) in out


def test_run__Runner____call____6(capsys):
    """It does not stop if ignore_exit_code is set."""
    pkgs = get_packages('bad', 'fine')
    runner = Runner([Command(raw_test_command, ignore_exit_code=True)], pkgs)
    runner()
    out, err = capsys.readouterr()
    out = remove_ansi_codes(out)
    assert 'Running {0} on {1[0]}'.format(raw_test_command, pkgs) in out
    assert 'Running {0} on {1[1]}'.format(raw_test_command, pkgs) in out
    assert out.strip().endswith('SUCCESS :-)')



import unittest


class Test_Bad(unittest.TestCase):
    """It fails."""

    def test_fine__1(self):
        self.assertTrue(False)


def test_suite():
    """Create test suite for `python setup.py test`."""
    return unittest.TestSuite([unittest.makeSuite(Test_Bad)])



import unittest


class Test_FineToo(unittest.TestCase):
    """It does not fail, too."""

    def test_fine__1(self):
        self.assertTrue(True)


def test_suite():
    """Create test suite for `python setup.py test`."""
    return unittest.TestSuite([unittest.makeSuite(Test_FineToo)])



from . import config
from .run import Runner
import argparse


def main(raw_args=None):
    """Console script entry point."""
    parser = argparse.ArgumentParser(
        description="poor man's integration testing")
    parser.add_argument(
        'cmds', metavar='cmd', default=['test'], nargs='*',
        help='Run command(s) defined in the configuration file. Each command '
             'is run on each package before proceeding with the next command. '
             '(default: "test")')
    parser.add_argument(
        '-c', '--config', dest='file', type=argparse.FileType('r'),
        default='toll.ini',
        help='ini-style file to read the configuration from')
    parser.add_argument(
        '--start-at', dest='start_at', type=str, default='',
        help='Skip over the packages in the config file listed before this'
             ' one. (It does a substring match to find the first package.)')

    args = parser.parse_args(raw_args)
    config_file = config.parsed_file(args.file)
    commands = config.commands(config_file, args.cmds)
    packages = config.packages(config_file)
    runner = Runner(commands, packages, start_at=args.start_at)
    return runner()



from .compat import ConfigParser


__all__ = ['parsed_file', 'packages', 'Command', 'commands']


def parsed_file(config_file):
    """Parse an ini-style config file."""
    parser = ConfigParser(allow_no_value=True)
    parser.readfp(config_file)
    return parser


def packages(config):
    """Return the packages defined in the config file as tuple."""
    _assert_section(config, 'packages')
    return tuple(config.options('packages'))


class Command:
    """Command which can be executed."""

    def __init__(self, command, precondition='', ignore_exit_code=False):
        self.command = command
        self.precondition = precondition
        self.ignore_exit_code = bool(ignore_exit_code)

    def __repr__(self):
        """Representation of this wrapper."""
        p = " if {0!r}".format(self.precondition) if self.precondition else ""
        i = " ignore-exit-code" if self.ignore_exit_code else ""
        return "<Command {0!r}{1}{2}>".format(self.command, p, i)

    def __eq__(self, other):
        """Compare two command instances."""
        if not isinstance(other, Command):
            return False
        return (self.command == other.command and
                self.precondition == other.precondition and
                self.ignore_exit_code == other.ignore_exit_code)

    def __ne__(self, other):
        """Own __eq__ requires own __ne__."""
        return not self.__eq__(other)


def minus_to_underscore(string):
    """Replace all minus in `string` by underscores."""
    return string.replace('-', '_')


def commands(config, names):
    """Return the list of commands to run."""
    commands = {cmd: Command(**dict((minus_to_underscore(k), v)
                                    for k, v in config.items(cmd)))
                for cmd in config.sections()
                if cmd != 'packages'}
    try:
        return tuple(commands[x] for x in names)
    except KeyError as e:
        raise RuntimeError(
            'Section [commands] in the config file does not contain the '
            'key {.args[0]!r} you requested to execute.'.format(e))


def _assert_section(config, name):
    """Make sure that a section exists in the config file."""
    if name not in config.sections():
        raise RuntimeError(
            'Missing the section [{}] in the config file.'.format(name))



from __future__ import print_function

from .compat import get_terminal_size
from colorama import Fore, Style
import os
import subprocess


GREEN = Fore.GREEN + Style.BRIGHT
YELLOW = Fore.YELLOW + Style.BRIGHT
RED = Fore.RED + Style.BRIGHT


class Runner:
    """Run the tests."""

    def __init__(self, commands, packages, start_at=''):
        self.commands = commands
        self.packages = packages
        self.start_at = start_at
        self.skipping_mode = True

    def __call__(self):
        for command in self.commands:
            for package in self.packages:
                if not self.run(
                        command.command, command.precondition,
                        command.ignore_exit_code, package):
                    self.render_failure()
                    return False
        self.render_success()
        return True

    def run(self, cmd, precondition, ignore_exit_code, package):
        if self.skipping_mode:
            if self.start_at in package:
                self.skipping_mode = False
            else:
                self._draw_line('#', Fore.YELLOW)
                print(YELLOW + 'Skipping', cmd, YELLOW + 'on', package)
                return True
        cwd = os.getcwd()
        os.chdir(package)
        try:
            if self._precodition_is_met(precondition):
                self._draw_line('*', Fore.GREEN)
                print(GREEN + 'Running', cmd, GREEN + 'on', package)
                result = self._run_cmd(cmd)
                if ignore_exit_code:
                    return True
                else:
                    return result
            else:
                self._draw_line('#', Fore.YELLOW)
                print(YELLOW + 'Not running', cmd, YELLOW + 'on', package)
                print(YELLOW + 'Precondition', precondition, YELLOW + 'on',
                      package, YELLOW + 'not met.')
                return True
        finally:
            os.chdir(cwd)

    def _precodition_is_met(self, precondition):
        if not precondition:
            return True
        if self._run_cmd(precondition):
            return True
        return False

    def _run_cmd(self, cmd):
        process = subprocess.Popen(cmd.split())
        return process.wait() == 0

    def _draw_line(self, char, colour):
        length = get_terminal_size().columns
        print(colour + char * length)

    def render_failure(self):
        print(RED + 'FAILURE :-(')

    def render_success(self):
        print(GREEN + 'SUCCESS :-)')



"""poor man's integration testing"""

from setuptools import setup, find_packages
import os.path
import shutil


def project_path(*names):
    """Path to a file in the project."""
    return os.path.join(os.path.dirname(__file__), *names)

test_require = [
    'pytest >= 3',
]

try:
    import unittest.mock  # noqa
except ImportError:
    test_require.append('mock')

install_requires = [
    'colorama',
]

if not hasattr(shutil, 'get_terminal_size'):
    install_requires.append('backports.shutil_get_terminal_size')

setup(
    name='toll',
    version='4.1',
    install_requires=install_requires,
    extras_require={
        'test': test_require,
    },

    entry_points={
        'console_scripts': [
            'toll = toll.main:main'
        ],
    },

    author='Michael Howitz',
    author_email='icemac@gmx.net',
    license='MIT',
    url='https://github.com/icemac/toll',

    keywords='test testing offline integration multiple packages',
    classifiers="""\
Development Status :: 4 - Beta
Environment :: Console
Intended Audience :: Developers
License :: OSI Approved :: MIT License
Natural Language :: English
Operating System :: OS Independent
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: Implementation :: CPython
Programming Language :: Python :: Implementation :: PyPy
Topic :: Software Development :: Testing
Topic :: Software Development :: Quality Assurance
Topic :: Utilities
"""[:-1].split('\n'),
    description=__doc__.strip(),
    long_description='\n\n'.join(open(project_path(name)).read() for name in (
        'README.rst',
        'CHANGES.rst',
    )),
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
)
